<?php
/**
 * Created by PhpStorm.
 * User: enured
 * Date: 01/05/24
 * Time: 4:00 PM
 */

namespace IMATHUZH\Qfq\Core\Form;


use IMATHUZH\Qfq\Core\Database\Database;
use IMATHUZH\Qfq\Core\Helper\Logger;
use IMATHUZH\Qfq\Core\Helper\OnArray;
use IMATHUZH\Qfq\Core\Helper\Path;
use IMATHUZH\Qfq\Core\Store\Store;
use Ratchet\MessageComponentInterface;
use Ratchet\ConnectionInterface;


/**
 * Class Chat
 * @package qfq
 */
class Chat implements MessageComponentInterface {
    protected \SplObjectStorage $clients;
    protected string $logFile;
    protected array $clientInfo;
    protected $clientConnections;

    /**
     * @throws \UserFormException
     */
    public function __construct() {
        $this->clients = new \SplObjectStorage;
        $this->logFile = Path::absoluteWebsocketLogFile();
        $this->clientInfo = [];
    }

    public function onOpen(ConnectionInterface $conn): void {
        $this->clients->attach($conn);
        $this->clientConnections[$conn->resourceId] = $conn;
        $timestamp = date("Y-m-d H:i:s");
        Logger::logMessage($timestamp . " - " . "New connection! ({$conn->resourceId})", $this->logFile);
    }

    public function onMessage(ConnectionInterface $from, $msg): void {

        $data = json_decode($msg);
        // Catch heartbeat connection keeping
        if ($data && property_exists($data, 'type') && $data->type == 'heartbeat') {
            return;
        }

        if ($data && property_exists($data, 'type') && $data->type == 'config') {
            // Handle the configuration data
            foreach ($data->data as $key => $value) {
                $this->clientInfo[$from->resourceId][$key] = $value;
            }

            $this->updateReceiverIds($from);

            $timestamp = date("Y-m-d H:i:s");
            Logger::logMessage($timestamp . " - " . "Connection configured! ({$from->resourceId}) Config: " . $msg, $this->logFile);
            return;
        }

        $this->refreshRecordId($from, $data);

        // Refresh receiverIds if there exist a new client connection
        $this->updateReceiverIds($from);

        if (!empty($this->clientInfo[$from->resourceId])) {
            // send ping to client with given receiverId
            foreach ($this->clientInfo[$from->resourceId][CHAT_WEBSOCKET_RECEIVER_IDS] as $receiverId) {
                if (isset($this->clientConnections[$receiverId])) {
                    $this->clientConnections[$receiverId]->send($data->data);
                }
            }
        }
    }

    public function onClose(ConnectionInterface $conn): void {
        $this->clients->detach($conn);
        unset($this->clientConnections[$conn->resourceId]);
        unset($this->clientInfo[$conn->resourceId]);
        $timestamp = date("Y-m-d H:i:s");
        Logger::logMessage($timestamp . " - " . "Connection {$conn->resourceId} has disconnected", $this->logFile);
    }

    public function onError(ConnectionInterface $conn, \Exception $e): void {
        $timestamp = date("Y-m-d H:i:s");
        Logger::logMessage($timestamp . " - " . "An error has occurred: {$e->getMessage()}", $this->logFile);
        $conn->close();
    }

    private function updateReceiverIds($from): void {
        if (!isset($this->clientInfo[$from->resourceId][CHAT_WEBSOCKET_RECEIVER_IDS])) {
            $this->clientInfo[$from->resourceId][CHAT_WEBSOCKET_RECEIVER_IDS] = array();
        }

        foreach ($this->clients as $client) {
            $fromCreator = $this->clientInfo[$from->resourceId][CHAT_COLUMN_PID_CREATOR] ?? null;
            $clientCreator = $this->clientInfo[$client->resourceId][CHAT_COLUMN_PID_CREATOR] ?? null;

            $fromRecordId = $this->clientInfo[$from->resourceId][CHAT_COLUMN_XID] ?? null;
            $clientRecordId = $this->clientInfo[$client->resourceId][CHAT_COLUMN_XID] ?? null;

            $fromTopicId = $this->clientInfo[$from->resourceId][CHAT_COLUMN_CID_TOPIC] ?? null;
            $clientTopicId = $this->clientInfo[$client->resourceId][CHAT_COLUMN_CID_TOPIC] ?? null;

            $fromThreadId = $this->clientInfo[$from->resourceId][CHAT_COLUMN_CID_THREAD] ?? null;
            $clientThreadId = $this->clientInfo[$client->resourceId][CHAT_COLUMN_CID_THREAD] ?? null;

            $isDifferentClient = $from !== $client;
            $isNotInReceiverIds = !in_array($client->resourceId, $this->clientInfo[$from->resourceId][CHAT_WEBSOCKET_RECEIVER_IDS]);

            $isValidCreator = empty($fromGroupList) && $fromCreator === $clientCreator;
            $isValidRecordId = $fromRecordId === $clientRecordId;
            $isValidTopicId = $fromTopicId === $clientTopicId;

            if ($isDifferentClient && $isValidRecordId && $isNotInReceiverIds && $isValidTopicId) {
                $this->clientInfo[$from->resourceId][CHAT_WEBSOCKET_RECEIVER_IDS][] = $client->resourceId;
            }
        }
    }

    /** Create chat configuration
     *
     * @param array $formElement
     * @param Store $store
     * @return array
     * @throws \CodeException
     * @throws \DbException
     * @throws \UserFormException
     * @throws \UserReportException
     */
    public static function createChatConfig(array $formElement): array {
        $chatConfig = array();

        // [apId:rId,pId:pIdCreator,grIdStatus:xGrIdStatus,grIdGroup:grIdGroupList,...]
        $chatConfig[CHAT_DB_COLUMN_NAMES] = array (
            'id' => 'id',
            'xId' => 'xId',
            'pIdCreator' => 'pIdCreator',
            'pIdTeam' => 'pIdTeam',
            'cIdTopic' => 'cIdTopic',
            'cIdThread' => 'cIdThread',
            'cIdLastRead' => 'cIdLastRead',
            'xGrIdStatus' => 'xGrIdStatus',
            'type' => 'type',
            'emoticon' => 'emoticon',
            'reference' => 'reference',
            'message' => 'message',
            'username' => 'username',
            'created' => 'created'
        );

        $chatConfig[CHAT_COLUMN_MAP] = $formElement[CHAT_COLUMN_MAP] ?? '[]';

        OnArray::mapColumns($chatConfig[CHAT_DB_COLUMN_NAMES], $chatConfig[CHAT_COLUMN_MAP]);
        $chatConfig[CHAT_TABLE_NAME] = $formElement[CHAT_TABLE_NAME] ?? 'Chat';

        $chatConfig[CHAT_COLUMN_XID] = $formElement[CHAT_PARAMETER_XID] ?? 0;
        $chatConfig[CHAT_COLUMN_PID_CREATOR] = empty($formElement[CHAT_PARAMETER_PID_CREATOR]) ? 0 : $formElement[CHAT_PARAMETER_PID_CREATOR];
        $chatConfig[CHAT_COLUMN_PID_TEAM] = $formElement[CHAT_PARAMETER_PID_TEAM] ?? 0;
        $chatConfig[CHAT_COLUMN_USERNAME] = empty($formElement[CHAT_PARAMETER_USERNAME]) ? 'Anonym' : $formElement[CHAT_PARAMETER_USERNAME];
        $chatConfig[CHAT_COLUMN_X_GRID_STATUS] = $formElement[CHAT_PARAMETER_X_GRID_STATUS] ?? 0;
        $chatConfig[CHAT_COLUMN_CID_TOPIC] = $formElement[CHAT_PARAMETER_CID_TOPIC] ?? 0;
        $chatConfig[CHAT_PARAMETER_OPTION_DONE] = $formElement[CHAT_PARAMETER_OPTION_DONE] ?? 'off';
        $chatConfig[CHAT_PARAMETER_OPTION_REMINDER] = $formElement[CHAT_PARAMETER_OPTION_REMINDER] ?? 0;
        $chatConfig[CHAT_PARAMETER_THREAD] = $formElement[CHAT_PARAMETER_THREAD] ?? 0;

        return $chatConfig;
    }

    /** Fetch from client requested messages and return them as a json.
     *
     * @param $chatConfig
     * @param Database $db
     * @param string $messageIdList
     * @param string $loadMode
     * @return array
     * @throws \CodeException
     * @throws \DbException
     * @throws \UserFormException
     */
    public static function createChatJson($chatConfig, Database $db, string $messageIdList = '', string $loadMode = CHAT_LOAD_MODE_REFRESH): array {
        $json = array();
        $sqlLimit = 11;

        // Define LIMIT 10 for lazy loading (refresh) or LIMIT 1 for client ping as new sender message
        if ($loadMode === CHAT_LOAD_MODE_PING) {
            $sqlLimit = 1;
        }

        $dbColumnNames = $chatConfig[CHAT_DB_COLUMN_NAMES] ?? array();
        // Get all dones withou threadId
        $sqlDoneNoThread = 'SELECT ' . implode(', ', array_map(function($column) use ($dbColumnNames) {
                return $dbColumnNames[$column];
            }, [CHAT_COLUMN_PID_CREATOR]))
            . ' FROM ' . $chatConfig[CHAT_TABLE_NAME] . ' WHERE ' . $dbColumnNames[CHAT_COLUMN_CID_TOPIC] . ' = ? AND '
            . $dbColumnNames[CHAT_COLUMN_XID] . ' = ? AND ' . $dbColumnNames[CHAT_COLUMN_TYPE] . ' = ? AND ' . $dbColumnNames[CHAT_COLUMN_MESSAGE]
            . ' =  ? ORDER BY ' . $dbColumnNames[CHAT_COLUMN_ID];

        $sqlDoneNotThreadParams = [$chatConfig[CHAT_COLUMN_CID_TOPIC], $chatConfig[CHAT_COLUMN_XID], CHAT_COLUMN_TYPE_TAG, CHAT_TYPE_TAG_DONE];

        // Get all messages from database. without threadId
        $sqlAllNotThread = 'SELECT ' . implode(', ', array_map(function($column) use ($dbColumnNames) {
                return $dbColumnNames[$column];
            }, [CHAT_COLUMN_ID, CHAT_COLUMN_XID, CHAT_COLUMN_PID_CREATOR, CHAT_COLUMN_CID_TOPIC, CHAT_COLUMN_MESSAGE, CHAT_COLUMN_USERNAME, CHAT_COLUMN_CREATED]))
            . ' FROM ' . $chatConfig[CHAT_TABLE_NAME]
            . ' WHERE ' . $dbColumnNames[CHAT_COLUMN_XID] . ' = ? AND ' . $dbColumnNames[CHAT_COLUMN_CID_TOPIC] . ' = ? AND !FIND_IN_SET(' . $dbColumnNames[CHAT_COLUMN_ID]
            . ', ?) AND ' . $dbColumnNames[CHAT_COLUMN_TYPE] . ' = ? '
            . 'ORDER BY ' . $dbColumnNames[CHAT_COLUMN_CREATED] . ' DESC LIMIT ' . $sqlLimit;

        $sqlAllNotThreadParams = [$chatConfig[CHAT_COLUMN_XID], $chatConfig[CHAT_COLUMN_CID_TOPIC], $messageIdList, CHAT_COLUMN_TYPE_MESSAGE];

        $resultAllDone = $db->sql($sqlDoneNoThread, ROW_REGULAR, $sqlDoneNotThreadParams);
        $resultAll = $db->sql($sqlAllNotThread, ROW_REGULAR, $sqlAllNotThreadParams);

        // Get moreResults flag for client and remove the last one
        $flagMoreRecords = count($resultAll) === 11;
        if ($flagMoreRecords) {
            array_pop($resultAll);
        }

        $chatJson = array();

        if (!empty($resultAll)) {
            foreach ($resultAll as $message) {
                $chatJson[$message[$dbColumnNames[CHAT_COLUMN_ID]]] = $message;
            }
        }
        // Return last message id as value
        if (!empty($resultAll)) {
            $columnId = array_column($resultAll, $dbColumnNames[CHAT_COLUMN_ID]);
            $recordId = array_column($resultAll, $dbColumnNames[CHAT_COLUMN_XID]);
            $createdDate = array_column($resultAll, $dbColumnNames[CHAT_COLUMN_CREATED]);

            $value = max($columnId);
            $recordId = max($recordId);
            $createdDate = max($createdDate);
        }

        $json[CHAT_COLUMN_XID] = $recordId ?? 0;
        $json[CHAT_COLUMN_CREATED] = $createdDate ?? '';
        $json[FE_TYPE_CHAT] = $chatJson;
        $json[CHAT_DONE] = $resultAllDone;
        $json[CHAT_COLUMN_PID_CREATOR] = $chatConfig[CHAT_COLUMN_PID_CREATOR];
        $json[CHAT_DB_COLUMN_NAMES] = $dbColumnNames;
        $json[CHAT_FLAG_MORE_RECORDS] = $loadMode === CHAT_LOAD_MODE_PING ? null : $flagMoreRecords;

        return $json;
    }

    /** Save new message given from client with given chat configuration
     *
     * @param $chatConfig
     * @param $feMode
     * @param Store $store
     * @param Database $db
     * @return array
     * @throws \CodeException
     * @throws \DbException
     * @throws \UserFormException
     * @throws \UserReportException
     */
    public static function saveChatMessage($chatConfig, $feMode, Store $store, Database $db): array {
        $json = array();
        $chatJson = array();

        if ($feMode[API_CHAT_UPDATE_DISABLED]) {
            return $json;
        }

        // Get current value from client store
        $value = $store::getVar(CHAT_CLIENT_MESSAGE, STORE_CLIENT, SANITIZE_ALLOW_ALLBUT);

        $chatConfig = Chat::createChatConfig($chatConfig);
        $chatParams = [
            CHAT_COLUMN_XID => $chatConfig[CHAT_COLUMN_XID] ?? 0,
            CHAT_COLUMN_PID_CREATOR => $chatConfig[CHAT_COLUMN_PID_CREATOR] ?? 0,
            CHAT_COLUMN_CID_TOPIC => $chatConfig[CHAT_COLUMN_CID_TOPIC] ?? 0,
            CHAT_COLUMN_USERNAME => $chatConfig[CHAT_COLUMN_USERNAME] ?? 0,
            CHAT_COLUMN_X_GRID_STATUS => $chatConfig[CHAT_COLUMN_X_GRID_STATUS] ?? 0
        ];

        $dbColumnNames = $chatConfig[CHAT_DB_COLUMN_NAMES];

        // Insert new record for chat
        $sqlInsert = 'INSERT INTO ' . $chatConfig[CHAT_TABLE_NAME] . '(' . $dbColumnNames[CHAT_COLUMN_XID]  . ', '. $dbColumnNames[CHAT_COLUMN_PID_CREATOR]
            . ',' . $dbColumnNames[CHAT_COLUMN_X_GRID_STATUS] . ',' . $dbColumnNames[CHAT_COLUMN_CID_TOPIC] . ','
            . $dbColumnNames[CHAT_COLUMN_MESSAGE] . ',' . $dbColumnNames[CHAT_COLUMN_USERNAME] . ')'
            . ' VALUES (?,?,?,?,?,?)';

        $insertParams = [$chatParams[CHAT_COLUMN_XID], $chatParams[CHAT_COLUMN_PID_CREATOR], $chatParams[CHAT_COLUMN_X_GRID_STATUS], $chatParams[CHAT_COLUMN_CID_TOPIC], $value, $chatParams[CHAT_COLUMN_USERNAME]];
        $id = $db->sql($sqlInsert, ROW_REGULAR, $insertParams);

        // Get inserted record data for chat json
        $sql = 'SELECT ' . implode(', ', array_map(function($column) use ($dbColumnNames) {
                return $dbColumnNames[$column];
            }, [CHAT_COLUMN_ID, CHAT_COLUMN_XID, CHAT_COLUMN_PID_CREATOR, CHAT_COLUMN_CID_TOPIC, CHAT_COLUMN_MESSAGE, CHAT_COLUMN_USERNAME, CHAT_COLUMN_CREATED]))
            . ' FROM ' . $chatConfig[CHAT_TABLE_NAME]
            . ' WHERE ' . $dbColumnNames[CHAT_COLUMN_ID] . ' = ?';
        $sqlParams = [$id];

        $resultAll = $db->sql($sql, ROW_REGULAR, $sqlParams);

        if (!empty($resultAll)) {
            foreach ($resultAll as $message) {
                $chatJson[$message[$dbColumnNames[CHAT_COLUMN_ID]]] = $message;
            }
        }

        $json[CHAT_COLUMN_XID] = $recordId ?? 0;
        $json[CHAT_COLUMN_CREATED] = $createdDate ?? date("Y-m-d H:i:s");
        $json[FE_TYPE_CHAT] = $chatJson;
        $json[CHAT_COLUMN_PID_CREATOR] = $chatConfig[CHAT_COLUMN_PID_CREATOR];
        $json[CHAT_DB_COLUMN_NAMES] = $chatConfig[CHAT_DB_COLUMN_NAMES];
        $json[CHAT_FLAG_MORE_RECORDS] = null;

        return $json;
    }


    /** Optional function to compare two different comma separated lists
     * If one of the value matches, true will be returned.
     *
     * @param $groupList1
     * @param $groupList2
     * @return bool
     */
    public static function existsInGroup($groupList1, $groupList2): bool {
        $bool = false;

        // Convert strings to arrays
        $groupArray1 = explode(",", $groupList1);
        $groupArray2 = explode(",", $groupList2);

        $matchingIds = array_intersect($groupArray1, $groupArray2);

        if (!empty($matchingIds)) {
            $bool = true;
        }

        return $bool;
    }

    /** Function to refresh xId in client info if given
     * @param $from
     * @param $data
     * @return void
     */
    private function refreshRecordId($from, $data): void {
        if ($this->clientInfo[$from->resourceId][CHAT_COLUMN_XID] == 0 && $data->xId) {
            $this->clientInfo[$from->resourceId][CHAT_COLUMN_XID] = $data->xId;
        }
    }
}